﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Drawing;
using System.IO;

namespace Sudoku
{
    public enum AdvancedAlgorithms { Nishio, Crook, DancingLinks, Off };

    public enum SolveStates { RozsaSolve, UserSolve, GettingHints, WritingData };

    public class AlgorithmManager
    {
        private static AlgorithmManager AlgorithmManager_instance;
        public static AlgorithmManager Instance
        {
            get 
            {
                if (AlgorithmManager_instance == null)
                {
                    AlgorithmManager_instance = new AlgorithmManager();
                }
                return AlgorithmManager_instance;
            }
        }

        #region variables and properties

        Crook crookAlgorithm;
        SudokuHistory sudokuHistory;
        bool error, foundPattern;
        int currentAlgorithm;
        SudokuTable original;
        IntPair currentPattern;
        Algorithm[] algorithms;

        AdvancedAlgorithms advancedAlgorithm;
        public AdvancedAlgorithms AdvancedAlgorithm
        {
            get { return advancedAlgorithm; }
            set { advancedAlgorithm = value; }
        }

        private int solveDepth;
        public int SolveDepth
        {
            get { return solveDepth; }
            set { solveDepth = value; }
        }

        SolveStates solveState;
        public SolveStates SolveState
        {
            get { return solveState; }
            set { solveState = value; }
        }

        private SudokuTable table;
        public SudokuTable Table
        {
            get { return table; }
            set { table = value; }
        }

        public string Report
        {
            get { return sudokuHistory.GetReport(original.M); }
        }

        public bool ShowCandidateTablesInReport
        {
            get { return sudokuHistory.ShowCandidatesTable; }
            set { sudokuHistory.ShowCandidatesTable = value; }
        }

        public bool ShowSimpleResult
        {
            get { return sudokuHistory.ShowSimpleResult; }
            set { sudokuHistory.ShowSimpleResult = value; }
        }

        List<List<Pattern>> findings;
        public List<List<Pattern>> Findings
        {
            get { return findings; }
            set { findings = value; }
        }

        SudokuControl sudokuControl;
        public SudokuControl SudokuControl
        {
            get { return sudokuControl; }
            set { sudokuControl = value; }
        }

        #endregion

        public bool GetAlgorithmState(int algindex)
        {
            return algorithms[algindex].Enabled;
        }
        public void SetAlgorithmState(int algindex, bool state)
        {
            algorithms[algindex].Enabled = state;
        }

        public void SetTableSize(int size)
        {
            error = false;
            table = new SudokuTable(size);
            original = new SudokuTable(size);
            //algorithms[2] = new Intersection(size);
        }

        public void RestartTable()
        {
            sudokuHistory.ShowCandidatesTable = true;
            crookAlgorithm = new Crook();
            error = false;
            table = original.Clone() as SudokuTable;
            sudokuHistory.ResetHistory();
            solveDepth = 0;

            ResetAlgorithms();
            GetHints();

            sudokuControl.UpdateTable(table);
        }

        private AlgorithmManager()
        {
            crookAlgorithm = new Crook();
            solveState = SolveStates.WritingData;
            advancedAlgorithm = AdvancedAlgorithms.Off;
            error = false;
            table = new SudokuTable(3);
            original = new SudokuTable();
            findings = new List<List<Pattern>>();
            sudokuHistory = new SudokuHistory();

            algorithms = new Algorithm[9];
            algorithms[0] = new NakedSingle();
            algorithms[1] = new HiddenSingle();
            algorithms[2] = new Intersection(table.N);
            algorithms[3] = new NakedPair();
            algorithms[4] = new HiddenPair();
            algorithms[5] = new NakedTriple();
            algorithms[6] = new HiddenTriple();
            algorithms[7] = new NakedQuad();
            algorithms[8] = new HiddenQuad();

            for (int i = 0; i < 11; ++i)
            {
                findings.Add(new List<Pattern>());
            }
        }

        public void StartNewPuzzle()
        {
            sudokuHistory.ShowCandidatesTable = true; 
            crookAlgorithm = new Crook();
            solveState = SolveStates.WritingData;
            error = false;
            original = Table.Clone() as SudokuTable;
            sudokuHistory.ClearHistory();
            solveDepth = 1;
            ResetAlgorithms();

            sudokuControl.UpdateTable(table);
        }

        public void StartNewPuzzle(SudokuTable t)
        {
            sudokuHistory.ShowCandidatesTable = true;
            crookAlgorithm = new Crook();
            solveState = SolveStates.UserSolve;
            error = false;
            table = t;
            original = Table.Clone() as SudokuTable;
            sudokuHistory.ClearHistory();

            //A kezdetben adott adatok betoltese a history-be
            solveDepth = 0;
            for (int i = 0; i < original.M; ++i)
            {
                for (int j = 0; j < original.M; ++j)
                {
                    if (original.Cells[i, j].IsData())
                    {
                        HistoryEntry he = new HistoryEntry(new IntPair(i, j), original.Cells[i, j].Data,
                            new List<IntPair>(), new List<int>(), -1);
                        he.Text = "r" + (i + 1).ToString() + "c" + (j + 1).ToString() + "=" + SudokuCell.ConvertCellDataToString(original.Cells[i, j].Data, table.M) + " (Adat)";
                        sudokuHistory.AddEntry(he, false);
                    }
                }
            }
            solveDepth = 1;

            ResetAlgorithms();
            GetHints();
            sudokuControl.UpdateTable(table);
        }

        public bool openwww(string url)
        {
            try
            {
                using (StreamReader sr = new StreamReader(System.Net.WebRequest.Create(url).GetResponse().GetResponseStream()))
                {
                    string sor;
                    while ((sor = sr.ReadLine()) != null)
                    {
                        string[] st = sor.Split('"');
                        foreach (string str in st)
                        {
                            if (Table.ValidString(str))
                            {
                                table = new SudokuTable();
                                if (Table.LoadFromString(str))
                                {
                                    StartNewPuzzle(table);
                                }
                                break;
                            }
                        }
                    }
                }
                sudokuControl.UpdateTable(table);
            }
            catch (Exception ex)
            {
                return false;
            }
            return true;
        }

        #region Rozsa

        public int sudokuType;

        public void Solve()
        {
            //ha az adatok melysegeben vagyunk, lepjunk egyel tovabb
            if (solveDepth == 0) solveDepth = 1;

            //???
            sudokuType = 0;

            //ha a rejtvenyrol mar kideritettuk, hogy nem jo, akkor ne csinaljunk semmit
            if (!sudokuHistory.IsEmpty && sudokuHistory.LastEntry.AlgorithmNumber == 9 && solveDepth == 1) return;

            bool solved = false;
            Nishio nishio = new Nishio();
            error = false;
            ResetAlgorithms();
            RemoveShownPattern();

            do
            {
                foundPattern = true;
                while (!error && foundPattern && table.DataInTable != table.M * table.M)
                {
                    foundPattern = false;
                    for (int i = 0; i < algorithms.Length && !error; i++)
                    {
                        //ha valamelyik algoritmus talált mintát ugrjunk vissza az egyszerű megoldó algoritmusokra
                        //if (i > 0 && foundPattern) i = 0;

                        currentAlgorithm = i;

                        solveState = SolveStates.RozsaSolve;
                        algorithms[i].solve(Table);

                        for (int k = currentAlgorithm; k >= 0; --k)
                        {
                            if (findings[k].Count > 0)
                            {
                                ProcessQueue(i);
                                i = currentAlgorithm - 1;
                            }
                        }

                        if (foundPattern && currentAlgorithm > 2)
                        {
                            i = currentAlgorithm = 1;
                            foundPattern = false;
                        }
                        if (table.DataInTable == table.M * table.M) break;
                    }
                }

                //ha az egyszeru algoritmusok sikerrel jartak kilepunk a ciklusbol
                if (table.DataInTable != table.M * table.M)
                {
                    HistoryEntry he;
                    switch (advancedAlgorithm)
                    {
                        case AdvancedAlgorithms.Nishio:
                            he = nishio.Guess(error);
                            if (he == null) return;
                            sudokuHistory.AddEntry(he, false);

                            error = false;
                            ResetAlgorithms();
                            break;
                        case AdvancedAlgorithms.Crook:
                            //tipp lekérése a Crook algoritmustól
                            he = crookAlgorithm.Guess(error);

                            //ha Crook algoritmus elakadt kilépünk a ciklusból
                            if (he == null) return;
                            error = false;

                            //egyébként hozzáadjuk a történethez
                            sudokuHistory.AddEntry(he, false);
                            
                            ResetAlgorithms();
                            break;
                        case AdvancedAlgorithms.DancingLinks:
                            DancingLinks dl = new DancingLinks();
                            dl.Table = Table;
                            dl.Run();
                            Table = dl.Table;
                            solved = true;
                            break;
                        case AdvancedAlgorithms.Off:
                            solved = true;
                            break;
                    }
                }
                else
                {
                    solved = true;
                }

            } while (!solved);

            if (!solved) GetHints();
            SolveState = SolveStates.UserSolve;
            sudokuControl.UpdateTable(Table);
        }

        private void ProcessQueue(int queueIndex)
        {
            if (queueIndex < 0) return;
            currentAlgorithm = queueIndex;
            while (findings[queueIndex].Count != 0 && !error)
            {
                //ha szisztematikus bejaras miatt a minta mar feldolgozasra kerult nem csinalunk semmit
                if (!Table.Cells[findings[queueIndex][0].CandidatePosition.i, findings[queueIndex][0].CandidatePosition.j].IsData())
                {
                    error = ApplyPattern(findings[queueIndex][0]);
                    //???
                    if (queueIndex > sudokuType) sudokuType = queueIndex;

                    if (error)
                    {
                        return;
                    }
                }
                findings[queueIndex].RemoveAt(0);
            }

            if (!error && findings[0].Count > 0) ProcessQueue(0);
            if (!error && findings[1].Count > 0) ProcessQueue(1);
        }

        public bool ApplyPattern(Pattern p)
        {
            if (!error && (p.AlgorithmIndex < 2 || p.AlgorithmIndex == 9 || p.IsValid()))
            {
                //???
                if (sudokuType < p.AlgorithmIndex) sudokuType = p.AlgorithmIndex;

                HistoryEntry newEntry;
                CandidatesTable ct = null;
                if (p.AlgorithmIndex == 10 || sudokuHistory.IsEmpty || sudokuHistory.LastEntry.AlgorithmNumber != p.AlgorithmIndex)
                {
                    ct = new CandidatesTable(table);
                }
                error = !p.ApplyPattern(Table, out newEntry);
                newEntry.CandidatesTable = ct;
                sudokuHistory.AddEntry(newEntry, false);
                foundPattern = true;
            }

            return error;
        }

        #endregion

        #region Solve

        public void LoadCrookTable()
        {
            if (!sudokuHistory.IsEmpty && sudokuHistory.LastEntry.AlgorithmNumber == 9)
            {
                crookAlgorithm.LoadBackupedTable();
                GetHints();
                Table.HideShownPattern();
                sudokuControl.UpdateTable(Table);
            }
        }

        private void ResetAlgorithms()
        {
            findings = new List<List<Pattern>>();
            for (int i = 0; i < 11; ++i )
            {
                findings.Add(new List<Pattern>());
            }
        }

        public void GetHints()
        {
            bool empty = true;
            bool writingData = (solveState == SolveStates.WritingData);
            SolveState = SolveStates.GettingHints;

            findings[9].Clear();
            for (int i = 0; i < 9; ++i )
            {
                findings[i].Clear();
                algorithms[i].solve(Table);
                if (findings[i].Count > 0) empty = false;
            }

            findings[10].Clear();

            if (empty && Table.DataInTable != Table.M * Table.M)
            {
                findings[10].Add(crookAlgorithm.GetCrookHint());
            }

            if (writingData) SolveState = SolveStates.WritingData;
            else SolveState = SolveStates.UserSolve;
        }

        public void AddPattern(Pattern p)
        {
            //alacsony indexu algoritmusok egy mintat biztosan csak egyszer talalnak meg
            if (p.AlgorithmIndex < 2 || findings[p.AlgorithmIndex].Count == 0 || p.AlgorithmIndex == 9)
            {
                findings[p.AlgorithmIndex].Add(p);
            }
            //nagyobb index eseten le kell ellenorizni, hogy a minta mar szerepel-e a talalatok kozt
            else
            {
                for (int i = 0; i < findings[p.AlgorithmIndex].Count; ++i)
                {
                    if (findings[p.AlgorithmIndex][i].Equals(p))
                    {
                        return;
                    }
                }
                findings[p.AlgorithmIndex].Add(p);
            }
        }

        public bool ApplyPattern(int alg, int ind)
        {
            if (alg > 10) return false;

            HistoryEntry newEntry;
            CandidatesTable ct = null;
            if (alg == 10) crookAlgorithm.RecordMove(findings[alg][ind].CandidatePosition.i, 
                findings[alg][ind].CandidatePosition.j, findings[alg][ind].CandidateValue);
            if (alg == 10 || sudokuHistory.IsEmpty || sudokuHistory.LastEntry.AlgorithmNumber != alg)
            {
                ct = new CandidatesTable(table);
            }
            error = !findings[alg][ind].ApplyPattern(Table, out newEntry);
            newEntry.CandidatesTable = ct;
            sudokuHistory.AddEntry(newEntry, false);

            GetHints();
            sudokuControl.UpdateTable(Table);
            
            return error;
        }

        public void UndoLastPattern()
        {
            if (sudokuHistory.UndoIsAllowed)
            {
                sudokuHistory.RemoveLastEntry();
                GetHints();
                sudokuControl.UpdateTable(Table);
            }
        }

        public void RemoveCandidate(int i, int j, int c)
        {
            Table.RemoveCandidate(i, j, c);
            sudokuHistory.AddEntry(new HistoryEntry(new IntPair(i, j), c, -2), false);
        
            GetHints();
            sudokuControl.UpdateTable(Table);
        }

        public void WriteCandidate(int i, int j, int c)
        {
            if (Table.Cells[i, j].IsData()) return;
            
            bool crook = true;
            for (int alg = 0; alg < 2; ++alg)
            {
                foreach (Pattern p in findings[alg])
                {
                    if (p.CandidateValue == c && p.CandidatePosition.i == i && 
                        p.CandidatePosition.j == j) crook = false;
                }
            }
            if (crook) crookAlgorithm.RecordMove(i, j, c);

            CandidatesTable ct = new CandidatesTable(table);
            HistoryEntry he = MakeWriteEntry(i, j, c);
            he.CandidatesTable = ct;
            Table.setCellData(i, j, c);
            sudokuHistory.AddEntry(he, true);

            if (solveState == SolveStates.WritingData && sudokuHistory.LastEntry.AlgorithmNumber >= 0)
            {
                solveState = SolveStates.UserSolve;
            }

            if (sudokuHistory.LastEntry.AlgorithmNumber < 0) crookAlgorithm = new Crook();

            GetHints();
            sudokuControl.UpdateTable(Table);
        }

        private HistoryEntry MakeWriteEntry(int i, int j, int c)
        {
            string newPatternStrings = "";
            List<IntPair> rcc = new List<IntPair>();
            List<int> rc = new List<int>();

            for (int row = 0; row < Table.M; ++row)
            {
                if (Table.Cells[row, j].IsCandidate(c) && row != i)
                {
                    Table.RemoveCandidate(row, j, c);
                    newPatternStrings += algorithms[0].TestCell(Table, new IntPair(row, j), new IntPair(i, j), c, c) +
                        algorithms[1].TestCell(Table, new IntPair(row, j), new IntPair(i, j), c, c);
                    rcc.Add(new IntPair(row, j));
                    rc.Add(c);
                }
            }

            for (int col = 0; col < Table.M; ++col)
            {
                if (Table.Cells[i, col].IsCandidate(c) && col != j)
                {
                    Table.RemoveCandidate(i, col, c);
                    newPatternStrings += algorithms[0].TestCell(Table, new IntPair(i, col), new IntPair(i, j), c, c) + 
                        algorithms[1].TestCell(Table, new IntPair(i, col), new IntPair(i, j), c, c);
                    rcc.Add(new IntPair(i, col));
                    rc.Add(c);
                }
            }

            for (int x = i / Table.N * Table.N; i / Table.N == x / Table.N; ++x)
            {
                for (int y = j / Table.N * Table.N; j / Table.N == y / Table.N; ++y)
                {
                    if (Table.Cells[x, y].IsCandidate(c) && x != i && y != j)
                    {
                        Table.RemoveCandidate(x, y, c);
                        newPatternStrings += algorithms[0].TestCell(Table, new IntPair(x, y), new IntPair(i, j), c, c) +
                            algorithms[1].TestCell(Table, new IntPair(x, y), new IntPair(i, j), c, c);
                        rcc.Add(new IntPair(x, y));
                        rc.Add(c);
                    }
                }
            }

            for (int cand = 1; cand < Table.M + 1; ++cand)
            {
                if (Table.Cells[i, j].IsCandidate(cand) && c != cand)
                {
                    Table.RemoveCandidate(i, j, cand);
                    newPatternStrings += algorithms[1].TestCell(Table, new IntPair(i, j), new IntPair(-1, -1), cand, c);
                    rcc.Add(new IntPair(i, j));
                    rc.Add(cand);
                }
            }

            HistoryEntry he = new HistoryEntry(new IntPair(i, j), c, rcc, rc, -1);
            he.Text = newPatternStrings;
            return he;
        }

        public string TestCell(IntPair redCell, IntPair greenCell, int redCand, int greenCand)
        {
            SolveState = SolveStates.UserSolve;

            string txt = "";

            txt += algorithms[0].TestCell(Table, redCell, greenCell, redCand, greenCand);
            if (currentAlgorithm > 0) txt += algorithms[1].TestCell(Table, redCell, greenCell, redCand, greenCand);
            if (currentAlgorithm > 1) txt += algorithms[currentAlgorithm].TestCell(Table, redCell, greenCell, redCand, greenCand);

            SolveState = SolveStates.RozsaSolve;

            return txt;
        }

        #endregion

        #region Paint

        public void RemoveShownPattern()
        {
            //elozo minta eltuntetese
            for (int i = 0; i < table.M; ++i)
            {
                for (int j = 0; j < table.M; ++j)
                {
                    table.Cells[i, j].BackGroundColor = 3;
                    for (int c = 1; c <= table.M; ++c)
                    {
                        table.Cells[i, j].SetColor(c, 0);
                    }
                }
            }
        }

        public void ChangeShownPattern(int alg, int ind)
        {
            currentPattern.i = alg;
            currentPattern.j = ind;

            Findings[alg][ind].PaintPattern(Table);
            sudokuControl.UpdateTable(Table);
        }

        public Color GetCellBackgroundColor(int x, int y)
        {
            return Color.Violet;
        }
        public Color GetNumberNoteColor(int x, int y, int c)
        {
            return Color.Violet;
        }

        #endregion
    }
}
