﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Sudoku
{
    class Crook
    {
        private struct GuessOption
        {
            public int Candidate;
            public IntPair CandidateCell;
        }

        private int depth;
        public int Depth
        {
            get { return depth; }
        }

        Stack<List<GuessOption>> guessOptions;
        Stack<SudokuTable> backupTables;

        public Crook()
        {
            depth = 0;
            guessOptions = new Stack<List<GuessOption>>();
            backupTables = new Stack<SudokuTable>();
        }

        private void newLevel()
        {
            SudokuTable latestBackupTable = AlgorithmManager.Instance.Table.Clone() as SudokuTable;
            List<GuessOption> guessOptionsLevel = new List<GuessOption>();

            bool found = false;
            for (int cnumber = 2; cnumber <= latestBackupTable.M && !found; ++cnumber)
            {
                for (int i = 0; i < latestBackupTable.M && !found; ++i)
                {
                    for (int j = 0; j < latestBackupTable.M && !found; ++j)
                    {
                        if (!latestBackupTable.Cells[i,j].IsData() && latestBackupTable.Cells[i, j].CandidatesNumber == cnumber)
                        {
                            for (int c = 1; c <= latestBackupTable.M; ++c)
                            {
                                if (latestBackupTable.Cells[i, j].IsCandidate(c))
                                {
                                    GuessOption go = new GuessOption();
                                    go.Candidate = c;
                                    go.CandidateCell = new IntPair(i, j);
                                    guessOptionsLevel.Add(go);
                                }
                            }
                            found = true;
                        }
                    }
                }
            }

            ++depth;
            AlgorithmManager.Instance.SolveDepth = depth + 1;
            guessOptions.Push(guessOptionsLevel);
            backupTables.Push(latestBackupTable);
        }

        public void RecordMove(int i, int j, int c)
        {
            if (guessOptions.Count != 0 && guessOptions.Peek().Count != 0 && guessOptions.Peek()[0].CandidateCell.i == i &&
                guessOptions.Peek()[0].CandidateCell.j == j)
            {
                    guessOptions.Peek().Remove(guessOptions.Peek().Find( delegate(GuessOption go) { return go.Candidate == c; }) );
            }
            else if (guessOptions.Count == 0 || guessOptions.Peek().Count == 0 || guessOptions.Peek()[0].CandidateCell.i != i ||
                guessOptions.Peek()[0].CandidateCell.j != j)
            {
                SudokuTable backupTable = AlgorithmManager.Instance.Table.Clone() as SudokuTable;
                List<GuessOption> guessOptionsLevel = new List<GuessOption>();

                for (int cand = 1; cand <= backupTable.M; ++cand)
                {
                    if (backupTable.Cells[i, j].IsCandidate(cand) && cand != c)
                    {
                        GuessOption go = new GuessOption();
                        go.Candidate = cand;
                        go.CandidateCell = new IntPair(i, j);
                        guessOptionsLevel.Add(go);
                    }
                }

                ++depth;
                AlgorithmManager.Instance.SolveDepth = depth + 1;

                guessOptions.Push(guessOptionsLevel);
                backupTables.Push(backupTable);
            }
        }

        /// <summary>
        /// Visszaadja a Crook algoritmus szerinti következő naplóbejegyzést
        /// Null-al tér vissza, ha az összes lehetőséget végigpróbálta az algoritmus
        /// </summary>
        /// <param name="error">híváskor a munkatábla hibás-e</param>
        /// <returns></returns>
        public HistoryEntry Guess(bool error)
        {
            //ha nem volt hiba, akkor újabb tippre van szükség, azaz egyel mélyebb szintre kell lepni
            if (!error) newLevel();

            //ha hiba és nulla mélységben vagyunk, akkor a rejtvény nem megoldható
            if (error && guessOptions.Count == 0) return null;

            //ha az adott szinten az összes lehetőséget végigpróbáltuk akkor feljebb kell lépni
            while (guessOptions.Peek().Count == 0)
            {
                guessOptions.Pop();
                backupTables.Pop();
                --depth;
                AlgorithmManager.Instance.SolveDepth = depth + 1;
            }

            //a munkatábla visszaállítása
            AlgorithmManager.Instance.Table = backupTables.Peek().Clone() as SudokuTable;

            //a tippelés
            GuessOption currentTipp = guessOptions.Peek()[0];
            guessOptions.Peek().RemoveAt(0);

            //a naplóbejegyzés elkészítése
            List<IntPair> rcc;
            List<int> rc;
            GetCandidatesToEliminate(out rc, out rcc, currentTipp.CandidateCell, currentTipp.Candidate);

            HistoryEntry he = new HistoryEntry(new IntPair(currentTipp.CandidateCell.i, currentTipp.CandidateCell.j), 
                currentTipp.Candidate, rcc, rc, -1);
            he.AlgorithmNumber = 10;
            he.Depth = Depth;
            he.Text = "r" + (currentTipp.CandidateCell.i + 1).ToString() + "c" + (currentTipp.CandidateCell.j + 1).ToString() + 
                "=" + currentTipp.Candidate.ToString() + " (Crook)";

            AlgorithmManager.Instance.Table.setCellData(currentTipp.CandidateCell.i, currentTipp.CandidateCell.j, currentTipp.Candidate);
            AlgorithmManager.Instance.Table.Cells[currentTipp.CandidateCell.i, currentTipp.CandidateCell.j].Foundby = depth + 1;

            for (int k = 0; k < rc.Count; ++k)
            {
                AlgorithmManager.Instance.Table.RemoveCandidate(rcc[k].i, rcc[k].j, rc[k]);
            }

            return he;
        }

        private void GetCandidatesToEliminate(out List<int> rc, out List<IntPair> rcc, IntPair ip, int c)
        {
            rc = new List<int>();
            rcc = new List<IntPair>();

            SudokuTable t = AlgorithmManager.Instance.Table;
            int n = t.N;
            int m = t.M;

            for (int row = 0; row < m; ++row)
            {
                if (t.Cells[row, ip.j].IsCandidate(c) && row != ip.i)
                {
                    rcc.Add(new IntPair(row, ip.j));
                    rc.Add(c);
                }
            }

            for (int col = 0; col < m; ++col)
            {
                if (t.Cells[ip.i, col].IsCandidate(c) && col != ip.j)
                {
                    rcc.Add(new IntPair(ip.i, col));
                    rc.Add(c);
                }
            }

            for (int x = ip.i / n * n; ip.i / n == x / n; ++x)
            {
                for (int y = ip.j / n * n; ip.j / n == y / n; ++y)
                {
                    if (t.Cells[x, y].IsCandidate(c) && ip.i != x && ip.j != y)
                    {
                        rcc.Add(new IntPair(x, y));
                        rc.Add(c);
                    }
                }
            }

            for (int cand = 1; cand < m + 1; ++cand)
            {
                if (cand != c)
                {
                    rcc.Add(new IntPair(ip.i, ip.j));
                    rc.Add(cand);
                }
            }
        }

        /// <summary>
        /// Visszaolvassa az utolsó táblát, ahonnan van remény megoldásra
        /// </summary>
        public void LoadBackupedTable()
        {
            if (backupTables.Count != 0)
            {
                if (guessOptions.Peek().Count != 0)
                {
                    AlgorithmManager.Instance.Table = backupTables.Peek().Clone() as SudokuTable;
                }
                else
                {
                    while (guessOptions.Peek().Count == 0)
                    {
                        guessOptions.Pop();
                        AlgorithmManager.Instance.Table = backupTables.Pop();
                        --depth;
                        AlgorithmManager.Instance.SolveDepth = depth + 1;
                    }
                    AlgorithmManager.Instance.Table = backupTables.Peek().Clone() as SudokuTable;
                }
            }
        }

        /// <summary>
        /// tipp kérése az algoritmustól
        /// </summary>
        /// <returns>a Rózsa algoritmus szerinti következő minta</returns>
        public CrookPattern GetCrookHint()
        {
            int tipp_c = 0;
            IntPair tipp_cell = new IntPair();
            SudokuTable t = AlgorithmManager.Instance.Table;

            bool found = false;
            for (int cnumber = 2; cnumber <= t.M && !found; ++cnumber)
            {
                for (int i = 0; i < t.M && !found; ++i)
                {
                    for (int j = 0; j < t.M && !found; ++j)
                    {
                        if (!t.Cells[i, j].IsData() && t.Cells[i, j].CandidatesNumber == cnumber)
                        {
                            for (int c = 1; c <= t.M && !found; ++c)
                            {
                                if (t.Cells[i, j].IsCandidate(c))
                                {
                                    tipp_c = c;
                                    tipp_cell = new IntPair(i,j);
                                    found = true;
                                }
                            }
                        }
                    }
                }
            }

            List<int> gc = new List<int>();
            List<IntPair> gcc = new List<IntPair>();
            List<IntPair> rcc;
            List<int> rc;
            if (guessOptions.Count != 0 && guessOptions.Peek().Count != 0 && tipp_cell.i == guessOptions.Peek()[0].CandidateCell.i && 
                tipp_cell.j == guessOptions.Peek()[0].CandidateCell.j)
            {
                GuessOption go = guessOptions.Peek()[0];
                gcc.Add(go.CandidateCell);
                gc.Add(go.Candidate);
            }
            else
            {
                gc.Add(tipp_c);
                gcc.Add(tipp_cell);
            }

            GetCandidatesToEliminate(out rc, out rcc, gcc[0], gc[0]);

            return new CrookPattern(gc[0], gc, gcc, rc, rcc);
        }
    }
}
