using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using System.Collections;

namespace Sudoku
{
    public partial class SudokuControl : UserControl
    {
        public enum SelectionRectangleSize { Small, Large };

        private Brush[] brushes = { Brushes.Black, Brushes.DarkGreen, Brushes.Red, Brushes.LightYellow, Brushes.LightGray, Brushes.PaleGreen};

        public delegate void TableChangedEventType(object o, SudokuTableEventArgs e);
        public event TableChangedEventType TableChanged;
        public delegate void WarningEventType(object o, string message);
        public event WarningEventType WarningEvent;
       
        #region Type: IntPair; Variables: SelectedCells, table, wl, hl, ctrldown; Properties: N, M, Littlefontfamily, Bigfontfamily, Pencilmarks

        struct IntPairWithColor
        {
            public IntPairWithColor(int x, int y, Color col)
            {
                pair.i = x;
                pair.j = y;
                color = col;
            }
            public IntPair pair;
            public Color color;
        }

        IntPair selectedcell;
        public IntPair SelectedCell
        {
            get 
            {
                if (selectionSize == SelectionRectangleSize.Large)
                {
                    return selectedcell;
                }
                else
                {
                    return new IntPair(selectedcell.i / n, selectedcell.j / n);
                }
            }
            set { selectedcell = value; }
        }

        SelectionRectangleSize selectionSize;
        public SelectionRectangleSize SelectionSize
        {
            get { return selectionSize; }
            set 
            {
                if (selectionSize == value) return;
                selectionSize = value;
                if (selectionSize == SelectionRectangleSize.Large)
                {
                    selectedcell.i /= n;
                    selectedcell.j /= n;
                }
                else
                {
                    selectedcell.i *= n;
                    selectedcell.j *= n;
                }
            }
        }
        
        List<IntPairWithColor> highlightedcells;

        FontFamily bigfontfamily;
        public FontFamily Bigfontfamily
        {
            get { return bigfontfamily; }
            set { bigfontfamily = value; }
        }

        FontFamily littlefontfamily;
        public FontFamily Littlefontfamily
        {
            get { return littlefontfamily; }
            set { littlefontfamily = value; }
        }

        SudokuTable table;

        bool pencilmarks = false;
        public bool Pencilmarks
        {
            get { return pencilmarks; }
            set { pencilmarks = value; Refresh(); }
        }

        int m = 9;
        public int M
        {
            get { return m; }
            private set { m = value; }
        }

        int n = 3;
        public int N
        {
            get { return n; }
            set
            {
                if (value > 0)
                {
                    n = value;
                    m = n * n;
                    table = new SudokuTable(n);
                    UnSelectCell();
                    highlightedcells.Clear();
                    OnTableChanged(this, new SudokuTableEventArgs(table, false, 0));
                    GC.Collect();
                    Refresh();
                }
            }
        }

        int wl, hl;

        private Brush[] DataColors = { Brushes.Black, Brushes.DarkBlue, Brushes.DarkGreen, Brushes.DarkSlateGray, 
                                         Brushes.DarkRed, Brushes.DarkSeaGreen, Brushes.DodgerBlue, Brushes.DarkViolet,
                                         Brushes.DarkSlateGray, Brushes.Firebrick, Brushes.ForestGreen, Brushes.DarkOrange};

        #endregion

        public SudokuControl()
        {
            MyInit();
            InitializeComponent();
        }

        public void Reset(int size)
        {
            N = size;
        }

        public void Reset()
        {
            Reset(n);
        }

        private void MyInit()
        {
            this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
            this.SetStyle(ControlStyles.ResizeRedraw, true);
            this.SetStyle(ControlStyles.DoubleBuffer, true);
            this.SetStyle(ControlStyles.UserPaint, true);
            table = new SudokuTable(n);
            selectedcell = new IntPair(-1, -1);
            highlightedcells = new List<IntPairWithColor>();
            littlefontfamily = new FontFamily("Arial");
            bigfontfamily = new FontFamily("Arial");
        }

        public void UpdateTable(SudokuTable t)
        {
            table = t.Clone() as SudokuTable;
            OnTableChanged(this, new SudokuTableEventArgs(table, false, 0));
            Refresh();
        }

        public void SelectCell(int i, int j)
        {
            selectedcell.i = i;
            selectedcell.j = j;
            Refresh();
        }

        public void UnSelectCell()
        {
            selectedcell.i = -1;
            selectedcell.j = -1;
            Refresh();
        }

        public void HighlightReset()
        {
            highlightedcells.Clear();
            Refresh();
        }

        public void Highlight(int c, Color color)
        {
            highlightedcells.Clear();
            for (int i = 0; i < m; i++)
            {
                for (int j = 0; j < m; j++)
                {
                    if (table.Cells[i, j].IsCandidate(c) && !table.Cells[i, j].IsData())
                    {
                        highlightedcells.Add(new IntPairWithColor(i, j, color));
                    }
                }
            }
            Refresh();
        }

        protected void OnTableChanged(object o, SudokuTableEventArgs e)
        {
            //ha feliratkozott valaki
            if (TableChanged != null)
            {
                TableChanged(o, e);
            }
        }

        protected void OnWarningEvent(object o, string message)
        {
            //ha feliratkozott valaki
            if (WarningEvent != null)
            {
                WarningEvent(o, message);
            }
        }

        public void CopyImageToClipboard()
        {
            Bitmap bmp = new Bitmap(wl * m, hl * m);
            Graphics g = Graphics.FromImage(bmp);
            DrawSudokuTable(g, Color.White);//fekete-fehr
            Clipboard.SetDataObject(bmp);//bmp vglapra msolsa
        }

        public void DrawSudokuTable(Graphics g, Color color)
        {
            //g.FillRectangle(new SolidBrush(color), 0, 0, wl * m, hl * m);

            for (int x = 0; x < m; ++x)
            {
                for (int y = 0; y < m; ++y)
                {
                    g.FillRectangle(brushes[table.Cells[y,x].BackGroundColor], x * wl, y * hl, x * wl + wl, y * hl + hl);
                }
            }

            if (color != Color.White)
            {
                highlightedcells.ForEach(delegate(IntPairWithColor ipc)
                {
                    g.FillRectangle(new SolidBrush(ipc.color), wl * ipc.pair.j + 1, hl * ipc.pair.i + 1, wl - 1, hl - 1);
                });
            }
            StringFormat format = new StringFormat(StringFormatFlags.FitBlackBox);
            format.LineAlignment = StringAlignment.Center;
            format.Alignment = StringAlignment.Center;
            Font f = new Font(littlefontfamily, (hl <= 0) ? 1 : hl, GraphicsUnit.Pixel);
            Font cf = new Font(bigfontfamily, (((hl / n) <= 0) ? 1 : (hl / n)), GraphicsUnit.Pixel);
            RectangleF displayRectangle = new RectangleF();

            for (int i = 0; i < m; i++)
            {
                for (int j = 0; j < m; j++)
                {
                    displayRectangle.X = wl * j + 1;
                    displayRectangle.Y = hl * i + 1;
                    displayRectangle.Width = wl;
                    displayRectangle.Height = hl;
                    if (table.Cells[i, j].IsData())
                    {
                        g.DrawString(SudokuCell.ConvertCellDataToString(table.Cells[i, j].Data, m), f, DataColors[table.Cells[i,j].Foundby % 12], displayRectangle, format);
                    }
                    else
                    {
                        if (pencilmarks)
                        {
                            for (int k = 1; k <= m; k++)
                            {
                                if (table.Cells[i, j].IsCandidate(k))
                                {
                                    g.DrawString(SudokuCell.ConvertCellDataToString(k, m), cf, brushes[table.Cells[i, j].GetColor(k)], new RectangleF(wl * j + ((k - 1) % n) * wl / n, hl * i + 1 + ((k - 1) / n) * hl / n, wl / n, hl / n), format);
                                }
                            }
                        }
                    }
                }
            }
            for (int i = 0; i <= m; i++)
            {
                g.DrawLine(new Pen(((i % n == 0) ? Color.Black : Color.Gray), ((i % n == 0) ? 2 : 1)), 0, hl * i, wl * m, hl * i);
                g.DrawLine(new Pen(((i % n == 0) ? Color.Black : Color.Gray), ((i % n == 0) ? 2 : 1)), wl * i, 0, wl * i, hl * m);
            }
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            wl = Width / m;
            hl = Height / m;
            Graphics g = e.Graphics;
            DrawSudokuTable(g, Color.LightYellow);
            if (selectionSize == SelectionRectangleSize.Large)
            {
                ControlPaint.DrawBorder(g, new Rectangle(wl * selectedcell.j + 1, hl * selectedcell.i + 1, wl - 1, hl - 1), Color.Blue, 2, ButtonBorderStyle.Outset, Color.Blue, 2, ButtonBorderStyle.Outset, Color.Blue, 2, ButtonBorderStyle.Outset, Color.Blue, 2, ButtonBorderStyle.Outset);
            }
            else
            {
                ControlPaint.DrawBorder(g, new Rectangle(wl * selectedcell.j / n + 1, hl * selectedcell.i / n + 1, wl / n - 1, hl / n - 1), Color.Red, 2, ButtonBorderStyle.Outset, Color.Red, 2, ButtonBorderStyle.Outset, Color.Red, 2, ButtonBorderStyle.Outset, Color.Red, 2, ButtonBorderStyle.Outset);
            }
            base.OnPaint(e);
        }

        protected override void OnKeyPress(KeyPressEventArgs e)
        {
            int c = 0;
            if (selectedcell.i >= 0 && selectionSize == SelectionRectangleSize.Large)
            {
                if (SudokuCell.IsLegalCharacter(e.KeyChar, m))
                {
                    c = SudokuCell.ConvertCharToCellData(e.KeyChar);
                    if (table.Cells[selectedcell.i, selectedcell.j].IsCandidate(c))
                    {
                        if (table.Cells[selectedcell.i, selectedcell.j].IsData())
                        {
                            int d = table.Cells[selectedcell.i, selectedcell.j].Data;
                            table.AddCandidateToThreeHouse(selectedcell.i, selectedcell.j, d);   
                        }
                        table.DeleteCandidateFromThreeHouse(selectedcell.i, selectedcell.j, c);
                        table.setCellData(selectedcell.i, selectedcell.j, c);
                        table.Cells[selectedcell.i, selectedcell.j].Foundby = 0;
                        OnTableChanged(this, new SudokuTableEventArgs(table, true, c));
                    }
                    Refresh();
                }
            }
            else if (selectedcell.i >= 0 && selectionSize == SelectionRectangleSize.Small)
            {
                if (SudokuCell.IsLegalCharacter(e.KeyChar, m))
                {
                    c = SudokuCell.ConvertCharToCellData(e.KeyChar);
                    if (table.Cells[selectedcell.i / n, selectedcell.j / n].IsCandidate(c) && selectedcell.i % n * n + selectedcell.j % n + 1 == c)
                    {
                        table.RemoveCandidate(selectedcell.i / n, selectedcell.j / n, c);
                        OnTableChanged(this, new SudokuTableEventArgs(table, true, c));
                    }
                    Refresh();
                }
            }

            base.OnKeyPress(e);
        }

        protected override bool ProcessDialogKey(Keys keyData)
        {
            if (selectedcell.i >= 0)
            {
                int maxIndex = 0;
                if (selectionSize == SelectionRectangleSize.Large)
                {
                    maxIndex = m - 1;
                }
                else
                {
                    maxIndex = m * n - 1;
                }

                switch (keyData)
                {
                    case Keys.Down:
                        if (selectedcell.i < maxIndex)
                        {
                            SelectCell(selectedcell.i + 1, selectedcell.j);
                            if (selectionSize == SelectionRectangleSize.Small && table.Cells[selectedcell.i / n, selectedcell.j / n].IsData())
                                SelectionSize = SelectionRectangleSize.Large;
                            Refresh();
                        }
                        break;
                    case Keys.Up:
                        if (selectedcell.i > 0)
                        {
                            SelectCell(selectedcell.i - 1, selectedcell.j);
                            if (selectionSize == SelectionRectangleSize.Small && table.Cells[selectedcell.i / n, selectedcell.j / n].IsData())
                                SelectionSize = SelectionRectangleSize.Large;
                            Refresh();
                        }
                        break;
                    case Keys.Left:
                        if (selectedcell.j > 0)
                        {
                            SelectCell(selectedcell.i, selectedcell.j - 1);
                            if (selectionSize == SelectionRectangleSize.Small && table.Cells[selectedcell.i / n, selectedcell.j / n].IsData())
                                SelectionSize = SelectionRectangleSize.Large;
                            Refresh();
                        }
                        break;
                    case Keys.Right:
                        if (selectedcell.j < maxIndex)
                        {
                            SelectCell(selectedcell.i, selectedcell.j + 1);
                            if (selectionSize == SelectionRectangleSize.Small && table.Cells[selectedcell.i / n, selectedcell.j / n].IsData())
                                SelectionSize = SelectionRectangleSize.Large;
                            Refresh();
                        }
                        break;
                    case Keys.Delete:
                        if (table.Cells[selectedcell.i, selectedcell.j].IsData())
                        {
                            table.AddCandidateToThreeHouse(selectedcell.i, selectedcell.j, table.Cells[selectedcell.i, selectedcell.j].Data);
                            table.setCellData(selectedcell.i, selectedcell.j, 0);
                            OnTableChanged(this, new SudokuTableEventArgs(table, true, 0));
                            Refresh();
                        }
                        break;
                    case Keys.Tab:
                        if (selectionSize == SelectionRectangleSize.Large)
                        {
                            SelectionSize = SelectionRectangleSize.Small;
                        }
                        else
                        {
                            SelectionSize = SelectionRectangleSize.Large;
                        }
                        Refresh();
                        break;
                }
            }
            return base.ProcessDialogKey(keyData);
        }

        protected override void OnMouseDown(MouseEventArgs e)
        {
            int i = 0, j = 0, maxIndex = 0;
            if (selectionSize == SelectionRectangleSize.Large)
            {
                maxIndex = m;
                i = e.Y / hl;
                j = e.X / wl;
            }
            else 
            {
                maxIndex = n * m;
                i = e.Y * n / hl;
                j = e.X * n / wl;
            }
            if (i >= 0 && j >= 0 && i < maxIndex && j < maxIndex)
            {
                if (selectedcell.i != i || selectedcell.j != j)
                {
                    SelectCell(i, j);
                }
            }
            base.OnMouseDown(e);
        }
    }

}
